"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const scaffold_1 = require("@0xcert/scaffold");
const web3_error_parser_1 = require("@0xcert/web3-error-parser");
const types_1 = require("./types");
class Context {
    constructor(options) {
        this.web3 = options.web3;
        this.myId = options.myId;
        this.exchangeId = options.exchangeId || '0x';
        this.signMethod = this.getSignMethod(options.signMethod);
    }
    attach() {
        return __awaiter(this, void 0, void 0, function* () {
            this.myId = yield this.getAccountId(this.myId);
            return this;
        });
    }
    detach() {
        return __awaiter(this, void 0, void 0, function* () {
            this.exchangeId = null;
            this.myId = null;
            this.signMethod = null;
            this.web3 = null;
            return this;
        });
    }
    sign(data) {
        return __awaiter(this, void 0, void 0, function* () {
            return `${this.signMethod}:${yield this.signData(data)}`;
        });
    }
    query(resolver) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return {
                    result: yield resolver(),
                };
            }
            catch (error) {
                throw web3_error_parser_1.parseError(error);
            }
        });
    }
    mutate(resolver, from) {
        return __awaiter(this, void 0, void 0, function* () {
            from = yield this.getAccountId(from || this.myId);
            try {
                const multiplyer = 1.2;
                const obj = yield resolver();
                let gas = yield obj.estimateGas({ from });
                gas = new this.web3.utils.BN(gas * 1.1);
                const price = yield this.web3.eth.getGasPrice().then((p) => p * multiplyer);
                const gasLimit = yield this.web3.eth.getBlock('latest').then((b) => b.gasLimit);
                if (gas > gasLimit) {
                    throw 'Gas exceeds max allowed gas';
                }
                return yield new Promise((resolve, reject) => {
                    const promise = obj.send({ from, gas, price });
                    promise.once('receipt', (tx) => resolve({ id: tx.transactionHash }));
                    promise.once('error', reject);
                }).then((d) => {
                    return d;
                });
            }
            catch (error) {
                throw web3_error_parser_1.parseError(error);
            }
        });
    }
    transfer(data, from) {
        return __awaiter(this, void 0, void 0, function* () {
            from = yield this.getAccountId(from || this.myId);
            try {
                const multiplyer = 1.2;
                const { value, to } = data;
                const gas = yield this.web3.eth.estimateGas({ from, to, value });
                const gasPrice = yield this.web3.eth.getGasPrice().then((p) => p * multiplyer);
                const gasLimit = yield this.web3.eth.getBlock('latest').then((b) => b.gasLimit);
                if (gas > gasLimit) {
                    throw 'Gas exceeds max allowed gas';
                }
                return yield new Promise((resolve, reject) => {
                    const promise = this.web3.eth.sendTransaction({ from, to, value, gas, gasPrice });
                    promise.once('transactionHash', (id) => resolve({ id }));
                    promise.once('error', reject);
                }).then((d) => {
                    return d;
                });
            }
            catch (error) {
                throw web3_error_parser_1.parseError(error);
            }
        });
    }
    getAccounts() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this.web3.eth.getAccounts();
            }
            catch (error) {
                throw web3_error_parser_1.parseError(error);
            }
        });
    }
    getAccountId(accountid) {
        return __awaiter(this, void 0, void 0, function* () {
            const accounts = yield this.getAccounts();
            if (accountid === undefined) {
                accountid = accounts[0];
            }
            if (accounts.indexOf(accountid) === -1) {
                throw new scaffold_1.ConnectorError(scaffold_1.ConnectorIssue.INVALID_MAKER_ID);
            }
            return accountid;
        });
    }
    getSignMethod(signMethod) {
        if (signMethod === undefined) {
            signMethod = types_1.SignMethod.ETH_SIGN;
        }
        if ([0, 1, 2].indexOf(signMethod) === -1) {
            throw new scaffold_1.ConnectorError(scaffold_1.ConnectorIssue.SIGNATURE_UNKNOWN, `Unknown signature method ${signMethod}`);
        }
        return signMethod;
    }
    signData(data) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                switch (this.signMethod) {
                    case types_1.SignMethod.ETH_SIGN:
                        return yield this.web3.eth.sign(data, this.myId);
                    case types_1.SignMethod.TREZOR:
                        return '';
                    case types_1.SignMethod.EIP712:
                        return '';
                    default:
                        return null;
                }
            }
            catch (error) {
                throw new scaffold_1.ConnectorError(scaffold_1.ConnectorIssue.SIGNATURE_FAILED, error);
            }
        });
    }
}
exports.Context = Context;
//# sourceMappingURL=context.js.map